;;;; -*- Mode: Lisp; indent-tabs-mode: nil -*-
;;;
;;; --- Wrapper over lfp_spawn(3)
;;;

(in-package :iolib.os)

(defclass process ()
  ((pid    :initarg :pid :reader process-pid)
   (stdin  :initform nil :reader process-stdin)
   (stdout :initform nil :reader process-stdout)
   (stderr :initform nil :reader process-stderr)))

(defmethod initialize-instance :after ((process process) &key stdin stdout stderr)
  (with-slots ((in stdin) (out stdout) (err stderr))
      process
    (setf in  (and stdin (make-instance 'iolib.streams:dual-channel-gray-stream :fd stdin))
          out (and stdout (make-instance 'iolib.streams:dual-channel-gray-stream :fd stdout))
          err (and stderr (make-instance 'iolib.streams:dual-channel-gray-stream :fd stderr)))))

(defmethod close ((process process) &key abort)
  (with-slots (pid stdin stdout stderr)
      process
    (when stdin  (close stdin :abort abort))
    (when stdout (close stdout :abort abort))
    (when stderr (close stderr :abort abort))
    (isys:waitpid pid (if abort isys:wnohang 0))
    (setf pid nil stdin nil stdout nil stderr nil)))

(defmethod print-object ((o process) s)
  (print-unreadable-object (o s :type t :identity t)
    (format s "~S ~S" :pid (process-pid o))))

(defmacro with-lfp-spawn-arguments ((attributes file-actions pid) &body body)
  (with-gensyms (spawnattr-initialized-p file-actions-initialized-p)
    `(with-foreign-objects ((,attributes 'lfp-spawnattr-t)
                            (,file-actions 'lfp-spawn-file-actions-t)
                            (,pid 'pid-t))
       (let (,spawnattr-initialized-p
             ,file-actions-initialized-p)
         (unwind-protect
              (progn
                (lfp-spawnattr-init ,attributes)
                (setf ,spawnattr-initialized-p t)
                (lfp-spawn-file-actions-init ,file-actions)
                (setf ,file-actions-initialized-p t)
                ,@body)
           (when ,spawnattr-initialized-p
             (lfp-spawnattr-destroy ,attributes))
           (when ,file-actions-initialized-p
             (lfp-spawn-file-actions-destroy ,file-actions)))))))

(defun allocate-argv (argv argc program arglist)
  ;; copy program name
  (setf (mem-aref argv :pointer 0)
        (foreign-string-alloc program))
  ;; copy program arguments
  (loop :for i :from 1
        :for arg :in arglist :do
        (setf (mem-aref argv :pointer i)
              (foreign-string-alloc arg)))
  ;; final null pointer
  (setf (mem-aref argv :pointer (1- argc)) (null-pointer)))

(defvar *default-shell* "/bin/sh")

(defun find-program (program)
  (cond
    ((eql :shell program)
     (or (nth-value 6 (isys:getpwuid (isys:getuid)))
         *default-shell*))
    (t
     (file-path-namestring program))))

(defmacro with-argv (((arg0 argv) program arguments) &body body)
  (with-gensyms (argc)
    `(let ((,program (find-program ,program))
           (,argc (+ 2 (length ,arguments))))
       (with-foreign-object (,argv :pointer ,argc)
         (unwind-protect
              (progn
                (allocate-argv ,argv ,argc ,program ,arguments)
                (let ((,arg0 (mem-ref ,argv :pointer)))
                  ,@body))
           (deallocate-null-ended-list ,argv))))))

(defun redirect-one-stream (stream fd file-actions &optional mode close-old-fd)
  (declare (notinline lfp-spawn-file-actions-addopen
                      lfp-spawn-file-actions-adddup2
                      lfp-spawn-file-actions-addclose))
  (flet ((dup-from-path (path)
           (lfp-spawn-file-actions-addopen file-actions fd path 0 mode))
         (dup-from-fd (oldfd)
           (lfp-spawn-file-actions-adddup2 file-actions oldfd fd)
           (when close-old-fd
             (lfp-spawn-file-actions-addclose file-actions oldfd))))
    (etypecase stream
      ((eql t) nil)
      ((or string file-path pathname)
       (dup-from-path (file-path-namestring stream)))
      ((eql :null)
       (dup-from-path "/dev/null"))
      (unsigned-byte
       (dup-from-fd stream))
      (iolib.streams:dual-channel-fd-mixin
       (dup-from-fd (iolib.streams:fd-of stream)))
      (null
       (lfp-spawn-file-actions-addclose file-actions fd)))))

(defun redirect-to-pipes (fd file-actions keep-write-fd)
  (declare (notinline isys:pipe))
  (multiple-value-bind (pipe-parent pipe-child)
      (isys:pipe)
    (when keep-write-fd (rotatef pipe-parent pipe-child))
    (lfp-spawn-file-actions-adddup2 file-actions pipe-parent fd)
    (lfp-spawn-file-actions-addclose file-actions pipe-parent)
    (lfp-spawn-file-actions-addclose file-actions pipe-child)
    (values pipe-parent pipe-child)))

(defun setup-redirections (file-actions stdin stdout stderr)
  (let (infd infd-child outfd outfd-child errfd errfd-child)
    (if (eql :pipe stdin)
        (setf (values infd infd-child)
              (redirect-to-pipes +stdin+ file-actions t))
        (redirect-one-stream stdin +stdin+ file-actions isys:o-rdonly))
    (if (eql :pipe stdout)
        (setf (values outfd outfd-child)
              (redirect-to-pipes +stdout+ file-actions nil))
        (redirect-one-stream stdout +stdout+ file-actions isys:o-wronly))
    (cond
      ((and stdout (eql :stdout stderr))
       (redirect-one-stream +stdout+ +stderr+ file-actions 0 nil))
      ((eql :pipe stderr)
       (setf (values errfd errfd-child)
             (redirect-to-pipes +stderr+ file-actions nil)))
      (t
       (redirect-one-stream stderr +stderr+ file-actions isys:o-wronly)))
    (values infd infd-child outfd outfd-child errfd errfd-child)))

(defun close-fds (&rest fds)
  (dolist (fd fds)
    (when fd (isys:close fd))))

(defmacro with-redirections (((infd outfd errfd) (file-actions stdin stdout stderr))
                             &body body)
  (with-gensyms (infd-child outfd-child errfd-child)
    `(multiple-value-bind (,infd ,infd-child ,outfd ,outfd-child ,errfd ,errfd-child)
         (setup-redirections ,file-actions ,stdin ,stdout ,stderr)
       (unwind-protect-case ()
           (locally ,@body)
         (:always
          (close-fds ,infd-child ,outfd-child ,errfd-child))
         (:abort
          (close-fds ,infd ,outfd ,errfd))))))

;; program: :shell - the current user's shell
;;          file-path designator - a path
;; arguments: list
;; search: boolean. whether or not to search PROGRAM in $PATH when PROGRAM only names a file,
;;                  i.e., it's a relative path whose directory is NIL
;; environment: t - inherit environment
;;              nil - NULL environment
;;              alist - the environment to use
;; stdin, stdout, stderr:
;;         file-path designator - open file, redirect to it
;;         :null - redirect to /dev/null - useful because /dev/null doesn't exist on Windows
;;         file-descriptor designator(integer or stream) - file descriptor, redirecto to it
;;         :pipe - create pipe, redirect the child descriptor to one end and wrap the other end
;;                 into a stream which goes into PROCESS slot
;;         t - inherit
;;         nil - close
;; stderr: :stdout - close if stdout is NIL, otherwise use the same fd

(defun create-process (program arguments &key (search t) (environment t)
                       (stdin t) (stdout t) (stderr t)
                       ;; path uid gid effective
                       )
  (declare (notinline lfp-spawn lfp-spawnp))
  (with-lfp-spawn-arguments (attributes file-actions pid)
    (with-argv ((arg0 argv) program arguments)
      (with-c-environment (envp environment)
        (with-redirections ((infd outfd errfd) (file-actions stdin stdout stderr))
          (if search
              (lfp-spawnp pid arg0 argv envp file-actions attributes)
              (lfp-spawn  pid arg0 argv envp file-actions attributes))
          (make-instance 'process :pid (mem-ref pid 'pid-t)
                         :stdin infd :stdout outfd :stderr errfd))))))

(defun run-program (program &optional arguments &key (search t) environment)
  (flet ((slurp-stream-into-string (stream)
           (with-output-to-string (s)
             (loop :for c := (read-char stream nil nil)
                   :while c :do (write-char c s)))))
    (let ((process (create-process program arguments
                                   :search search
                                   :environment environment)))
      (values (process-wait process)
              (slurp-stream-into-string (process-stdout process))
              (slurp-stream-into-string (process-stderr process))))))

(defun process-wait (process)
  (isys:waitpid (process-pid process) 0))

(defun process-kill (process signum)
  (isys:kill (process-pid process) signum))
